/*
 *  Petit Decoder Program (based on Head Tail decoder 2019-12-08 )
 *
 *  Pin assign 
 *    (OC0A)(TPID)PB0 1   6 PB3(RST)
 *                GND 2   5 Vcc
 *    (OC0B)(TPIC)PB1 3   4 PB2   
 *
 *   PB0 = OUTA(OUT1) 
 *   PB1 = OUTB(OUT2)   前進時HI
 *   PB2 = DCC command input アナログ運転時にHIにする。
 * 
 rev 0 2020-4-21 arduino attiny108bitDuino10), 1MHz(Internal)
           writer = AVRISP

       2020-5-18 Check at FL12 type Petit Decoder
       2020-6-5 add comments
*/
           
 //////////////////////////////////////////////////////
//
//   User setting area
//
//   (1) Select DCC loco address mode
//    On range 100-127
//    It is confusing 
//    which mode to choose 
//    depending on the vendor of the command station
//    #define SHORTADRS // select address mode
    #define LONGADRS   //
//  
#define  LOCOADRS   283 // <==== set your loco address
//
//
//
//////////////////////////////////////////////////////
//
// Define each bit from loco address.
//
#define B0  ( LOCOADRS                                                                                           % 2)
#define B1  ((LOCOADRS                                                                                 -B0)/   2 % 2)
#define B2  ((LOCOADRS                                                                            -2*B1-B0)/   4 % 2)
#define B3  ((LOCOADRS                                                                       -4*B2-2*B1-B0)/   8 % 2)
#define B4  ((LOCOADRS                                                                  -8*B3-4*B2-2*B1-B0)/  16 % 2)
#define B5  ((LOCOADRS                                                            -16*B4-8*B3-4*B2-2*B1-B0)/  32 % 2)
#define B6  ((LOCOADRS                                                      -32*B5-16*B4-8*B3-4*B2-2*B1-B0)/  64 % 2)
#define B7  ((LOCOADRS                                                -64*B6-32*B5-16*B4-8*B3-4*B2-2*B1-B0)/ 128 % 2)
#define B8  ((LOCOADRS                                         -128*B7-64*B6-32*B5-16*B4-8*B3-4*B2-2*B1-B0)/ 256 % 2)
#define B9  ((LOCOADRS                                  -256*B8-128*B7-64*B6-32*B5-16*B4-8*B3-4*B2-2*B1-B0)/ 512 % 2)
#define B10 ((LOCOADRS                           -512*B9-256*B8-128*B7-64*B6-32*B5-16*B4-8*B3-4*B2-2*B1-B0)/1024 % 2)
#define B11 ((LOCOADRS                  -1024*B10-512*B9-256*B8-128*B7-64*B6-32*B5-16*B4-8*B3-4*B2-2*B1-B0)/2048 % 2)
#define B12 ((LOCOADRS         -2048*B11-1024*B10-512*B9-256*B8-128*B7-64*B6-32*B5-16*B4-8*B3-4*B2-2*B1-B0)/4096 % 2)
#define B13 ((LOCOADRS-4096*B12-2048*B11-1024*B10-512*B9-256*B8-128*B7-64*B6-32*B5-16*B4-8*B3-4*B2-2*B1-B0)/8192 % 2)
//
//
#include <avr/io.h>
#define cbi(addr,bit)     addr &= ~(1<<bit) // cbi()    addrのbit番目をLowに
#define sbi(addr,bit)     addr |=  (1<<bit) // sbi()    addrのbit番目をHighに
#define sigIsHigh bit_is_set(PINB,PINB2)
#define sigIsLow bit_is_clear(PINB,PINB2)
//
// Read a BIT of the DCC command
//
unsigned char readBit() {
  for(;sigIsLow!=0;); // Suspennding while the shignal is low
  for(short j=0;j<12;j++) cbi(PORTB,PINB3);  // 59μS
  if(sigIsLow!=0) return 1;// if the signal is low then BIT is "1" and return funnction
  for(;sigIsLow==0;); //　Sispending while the signal is high
  return 0;
}
//
// Add noize protection (2017-11-27)
//
unsigned char findDCCData() { 
  unsigned char cnt=0;
  for(;sigIsLow!=0;)if(++cnt>200) return 0; // waiting while the signal is low(with timeout function)
  for(short j=0;j<8;j++) cbi(PORTB,PINB3);  // 40μS ?
  if(sigIsLow!=0) return 0;// if the signal is low, ノイズなのですぐ戻る
  for(short j=0;j<12;j++) cbi(PORTB,PINB3);  // 59μS
  if(sigIsLow!=0) return 1;// lowに戻っていたらデータなのですぐ戻る
  for(cnt=0;sigIsLow==0;)if(++cnt>10) return 0; // highのうちは待つが、タイアウトあり
  return 0;
}
//
// ノイズを1で返さないようにしたPWM検知バージョン(アナログ運転時のみ使う)
//
unsigned char findAnalogPulse() {
  unsigned char cnt=0;
  for(;sigIsLow!=0;)if(++cnt>100) return 0; // lowのうちは待つが、タイアウトあり
  for(short j=0;j<10;j++) cbi(PORTB,PINB3);  // 40μS ? 8を5にした！！！ かえっておかしくなったので10にする
  if(sigIsLow!=0) return 0;// lowに戻っていたらノイズなのですぐ戻る
  return 1; //それ以上は１　ＤＣＣパルスとする
}
/*
** mS単位のウエイト関数
*/
void wait_ms(short time) {
  for(;time!=0;time--) for(short j=0;j<200;j++) cbi(PORTB,PINB3); //ダミー処理 500mSでチューニング済
}
//
// アナログモードの処理
//
void decideAnalogModeDireciton() {
  unsigned char a=50;
resetDirection:
  for(;;) {
    // パルスが入る方
    if(findAnalogPulse()!=0)
      /* 一度パルスが検知されたら、なかなか抜けないようにしている */
      {sbi(PORTB,PINB1);  cbi(PORTB,PINB0);
        for(;;) {
          // wait_ms(100);
          if(findAnalogPulse()!=0)
            a=50;
          else {
            if (--a==0) goto resetDirection ;
          }
        }
      }
    // LOで固定されている方
    else{
      sbi(PORTB,PINB0);    cbi(PORTB,PINB1);
      wait_ms(100); //ウエイトで調べる　動力車を置くと　こちらのサイレンス側によく戻ることがわかる。
    }
  }
}
void analogModeOld() {
  for(;;) {
    if(sigIsLow!=0)
    {sbi(PORTB,PINB0);  cbi(PORTB,PINB1);}
    else
    {sbi(PORTB,PINB1); cbi(PORTB,PINB0);}
  }
}
//
// 連続した１が何ビット続くかカウントする
//
unsigned char checkPriambleLength() {
  unsigned char count;
  for (count=0;readBit()!=0;count++);
  return count;
} 
//
// オフにするのだけ関数
//
void turnoff(void) { 
  if(TCCR0A!=0) {
    OCR0BL=0;
    for(OCR0AL=0;OCR0AL!=0xff;OCR0AL++) 
      {OCR0BL++;wait_ms(1);}
    TCCR0A=0;
  } 
}
//
//
void setup() {

/*int main(void) {*/
  unsigned char sum,cmd,dir=0,F0BIT=0;
  
  DDRB= 0B00000011;   // PB0,PB1を出力にする PB2はパルスのセンス回路に接続する
  sbi(PUEB,PINB2);  // PB2をプルアップにしてみる
  
  // まずは0.1秒全点灯
//  sbi(PORTB,PINB1);wait_ms(100);cbi(PORTB,PINB1);  //片方点灯させて様子を見るとき
//  sbi(PORTB,PINB0);wait_ms(100);cbi(PORTB,PINB0);  //片方点灯させて様子を見るとき
  wait_ms(2000);    // 通電されてからの待ち時間を延ばすD101がおそそう 100=>250mS=>2000mS
  
  // 4回連続してDCCデータ1があればDCC、50回さがしてなければアナログに移行
  //  2017-11-28 ここはもういじらないこと！
  //  2019-12-8 DCCにいかない場合がある
  sum=0;  // 変数をまたがりした  
checkAnalog:  
  if(++sum>50)  decideAnalogModeDireciton(); // アナログモード　
  if(findDCCData()==0) goto checkAnalog;    //一回だけだとアナログでも誤動作する
  if(findDCCData()==0) goto checkAnalog;  
  if(findDCCData()==0) goto checkAnalog;  
  // 3回にへらした if(findDCCData()==0) goto checkAnalog; 

  //for(;;) {
  //    { sbi(PORTB,PINB1);wait_ms(10); cbi(PORTB,PINB1);}
  //    { sbi(PORTB,PINB0);wait_ms(10); cbi(PORTB,PINB0);}
  //    { sbi(PORTB,PINB1);wait_ms(10); cbi(PORTB,PINB1);}
  //    { sbi(PORTB,PINB0);wait_ms(10); cbi(PORTB,PINB0);}
  //  }
  // ここからはDCCと判断した後の処理である。
  
  OCR0AL=OCR0BL=0xff; // PWMの出力をゼロに初期化する
  
  //
  // Waveform Generation Modes
  //  WGM=0101 Fast PWM 9bit TOP=0xff
  //
  // TCCR0A
  // 7:COM0A1  = 1 use
  //   COM0A0 = 1 
  // 7:COM0B1  = 1 use
  //   COM0B0 = 1 
  // -
  // -
  // WGM01  0 (8bit設定)
  // WGM00   1  (8bit設定)
  //  TCCR0A = 0b11000001; // COMOAは利用,COMはオフ　ボトムでオン、マッチでオフ - コマンドが来るまでは、設定しない
  // TCCR0B
  // ICNC=0
  // ICES=0
  // -
  // WGM03=0
  // WGM02=1
  // CS02 = 0 4mS周期
  // CS01 = 1
  // CS00 = 0
  TCCR0B = 0b00001010;

  for(;;) {
    // プリアンブルチェック
checkpriamble:
    // プリアンブルカウント,エンドサーチ
//    if(readBit()==0) goto checkpriamble; // ゼロなら先頭へ
//    if(readBit()==0) goto checkpriamble; // ゼロなら先頭へ
//    if(readBit()==0) goto checkpriamble; // ゼロなら先頭へ
//    if(readBit()==0) goto checkpriamble; // ゼロなら先頭へ
//    if(readBit()==0) goto checkpriamble; // ゼロなら先頭へ
//    if(readBit()==0) goto checkpriamble; // ゼロなら先頭へ
//    if(readBit()==0) goto checkpriamble; // ゼロなら先頭へ
//    if(readBit()==0) goto checkpriamble; // ゼロなら先頭へ
//    if(readBit()==0) goto checkpriamble; // ゼロなら先頭へ
//waitend:
//    if(readBit()!=0) goto waitend;    // プリアンブルの終了を待つ
    
    if(checkPriambleLength()<8) goto checkpriamble;
    
#ifdef  SHORTADRS
    // adrs1 == 03 -> 0000 0011
    if(readBit()!=0) goto checkpriamble;    //期待値は0
    if(readBit()!=B6) goto checkpriamble;   //期待値はB6
    if(readBit()!=B5) goto checkpriamble;   //期待値はB5
    if(readBit()!=B4) goto checkpriamble;   //期待値はB4
    if(readBit()!=B3) goto checkpriamble;   //期待値はB3
    if(readBit()!=B2) goto checkpriamble;   //期待値はB2
    if(readBit()!=B1) goto checkpriamble;   //期待値はB1
    if(readBit()!=B0) goto checkpriamble;   //期待値はB0
    if(readBit()) goto checkpriamble;     // バイト間ギャップの確認
#endif
#ifdef  LONGADRS //  460 => 000001 11001100
    // adrs1 == C1 -> 1100 0001
    if(readBit()==0) goto checkpriamble;    //期待値は1 固定値
    if(readBit()==0) goto checkpriamble;    //期待値は1 固定値
    if(readBit()!=B13) goto checkpriamble;    //期待値は0
    if(readBit()!=B12) goto checkpriamble;    //期待値は0
    if(readBit()!=B11) goto checkpriamble;    //期待値は0
    if(readBit()!=B10) goto checkpriamble;    //期待値は0
    if(readBit()!=B9) goto checkpriamble;   //期待値は0
    if(readBit()!=B8) goto checkpriamble;   //期待値は1
    if(readBit()) goto checkpriamble;     // バイト間ギャップの確認

    // adrs2 == CC -> 1100 1100 
    if(readBit()!=B7) goto checkpriamble;   //期待値は1
    if(readBit()!=B6) goto checkpriamble;   //期待値は1
    if(readBit()!=B5) goto checkpriamble;   //期待値は0
    if(readBit()!=B4) goto checkpriamble;   //期待値は0
    if(readBit()!=B3) goto checkpriamble;   //期待値は1
    if(readBit()!=B2) goto checkpriamble;   //期待値は1
    if(readBit()!=B1) goto checkpriamble;   //期待値は0
    if(readBit()!=B0) goto checkpriamble;   //期待値は0
    if(readBit()) goto checkpriamble;     // バイト間ギャップの確認
#endif

    // コマンド1の読み出し 3Fのはず-> 0011 1111 ,ただし先頭が1なら　F0-F4コマンドであるかチェックする
    if(readBit()!=0) goto ckeckF0BIT;   //期待値は0なら3Fコマンドかチェック、1なら08x,09コマンドかのチェックにジャンプ
    // check3F:   
    if(readBit()!=0) goto checkpriamble;    //期待値は0
//    if(readBit()==0) goto checkpriamble;    //期待値は1
//    if(readBit()==0) goto checkpriamble;    //期待値は1
//    if(readBit()==0) goto checkpriamble;    //期待値は1
//    if(readBit()==0) goto checkpriamble;    //期待値は1
//    if(readBit()==0) goto checkpriamble;    //期待値は1
//    if(readBit()==0) goto checkpriamble;    //期待値は1
//    if(readBit()) goto checkpriamble;     // バイト間ギャップの確認
// 上記６ビット分は以下に置き換え
    if(checkPriambleLength()!=6) goto checkpriamble;
    cmd = readBit(); cmd = cmd << 1;
    cmd += readBit(); cmd = cmd << 1;
    cmd += readBit(); cmd = cmd << 1;
    cmd += readBit(); cmd = cmd << 1;
    cmd += readBit(); cmd = cmd << 1;
    cmd += readBit(); cmd = cmd << 1;
    cmd += readBit(); cmd = cmd << 1;
    cmd += readBit();
    if(readBit()) goto checkpriamble;     // バイト間ギャップの確認
    
    sum = readBit(); sum = sum << 1;
    sum += readBit(); sum = sum << 1;
    sum += readBit(); sum = sum << 1;
    sum += readBit(); sum = sum << 1;
    sum += readBit(); sum = sum << 1;
    sum += readBit(); sum = sum << 1;
    sum += readBit(); sum = sum << 1;
    sum += readBit();

#ifdef SHORTADRS
    if((LOCOADRS^0x3f^cmd^sum)!=0) goto checkpriamble;  
#endif
#ifdef LONGADRS
    if(((0xc0|(LOCOADRS/256))^(LOCOADRS%256)^0x3f^cmd^sum)!=0) goto checkpriamble; 
#endif
    //  F0で消灯の場合
    if(F0BIT==0){ turnoff();dir=0;goto checkpriamble; }
    // F0で点灯の場合
    if(cmd & 0x80) { // 11/23 ココのロジックを逆にした
      if(dir==1)  turnoff();      // 他がついていたらまず消灯する
      if(dir!=2) {      // 点灯を開始する
        dir=2;
        if(TCCR0A != 0b00110001) TCCR0A=0b00110001;
        for(OCR0BL=0xff;OCR0BL!=0;OCR0BL--) wait_ms(1);
      }
    }
    else {
      // 他がついていたらまず消灯する
      if(dir==2)  turnoff();
      // 点灯を開始する
      if(dir!=1) {
        dir=1;
        if(TCCR0A != 0b11000001) TCCR0A=0b11000001;
        for(OCR0AL=0xff;OCR0AL!=0;OCR0AL--) wait_ms(1);
      }
    }     
    // LEDの更新
    //if(lighton==1 && dir==1) {sbi(PORTB,PINB0);  cbi(PORTB,PINB1);}
    //if(lighton==1 && dir==2)  {sbi(PORTB,PINB1); cbi(PORTB,PINB0);}
    //if(lighton==0) {cbi(PORTB,PINB1); cbi(PORTB,PINB0);}
    goto checkpriamble;

//===============ライトの点灯、消灯処理=======================
ckeckF0BIT:
    cmd = readBit()+2; cmd = cmd << 1;  // 最初のビットは１　次から読み出す
    cmd += readBit(); cmd = cmd << 1;
    cmd += readBit(); cmd = cmd << 1;
    cmd += readBit(); cmd = cmd << 1;
    cmd += readBit(); cmd = cmd << 1;
    cmd += readBit(); cmd = cmd << 1;
    cmd += readBit();
    if(readBit()) goto checkpriamble;     // バイト間ギャップの確認
    sum = readBit(); sum = sum << 1;
    sum += readBit(); sum = sum << 1;
    sum += readBit(); sum = sum << 1;
    sum += readBit(); sum = sum << 1;
    sum += readBit(); sum = sum << 1;
    sum += readBit(); sum = sum << 1;
    sum += readBit(); sum = sum << 1;
    sum += readBit();
#ifdef SHORTADRS
    if((LOCOADRS^cmd^sum)!=0) goto checkpriamble;  // チェックサムを確認
#endif
#ifdef LONGADRS
    if(((0xc0|(LOCOADRS/256))^(LOCOADRS%256)^cmd^sum)!=0) goto checkpriamble;  // チェックサムを確認 ここは通っている
#endif
    if((cmd&0b11100000)==0b10000000) F0BIT = cmd&0b00010000; // F0-F4指定コマンド (8x　or 9X)
  } 
}
void loop() {
  // put your main code here, to run repeatedly:

}
// End of file
